# Copyright (C) 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import re
import sys

import pytest

from tests import utils

import virtinst.progress
from virtinst import Installer
from virtinst import Guest
from virtinst import log

# These are all functional tests
os.environ.pop("VIRTINST_TEST_SUITE", None)


class _URLTestData:
    """
    Class that tracks all data needed for a single URL test case.
    Data is stored in data/test_urls.ini
    """

    def __init__(
        self,
        name,
        url,
        detectdistro,
        testxen,
        testshortcircuit,
        kernelarg,
        kernelregex,
        skip_libosinfo,
    ):
        self.name = name
        self.url = url
        self.detectdistro = detectdistro
        self.arch = self._find_arch()
        self.kernelarg = kernelarg
        self.kernelregex = kernelregex
        self.skip_libosinfo = skip_libosinfo

        self.testxen = testxen

        # If True, pass in the expected distro value to getDistroStore
        # so it can short circuit the lookup checks. Speeds up the tests
        # and exercises the shortcircuit infrastructure
        self.testshortcircuit = testshortcircuit

    def _find_arch(self):
        if "i686" in self.url or "i386" in self.url or "i586" in self.url:
            return "i686"
        if "arm64" in self.url or "aarch64" in self.url:
            return "aarch64"
        if "ppc64el" in self.url or "ppc64le" in self.url:
            return "ppc64le"
        if "s390" in self.url:
            return "s390x"
        if "x86_64" in self.url or "amd64" in self.url:
            return "x86_64"
        return "x86_64"


testconn = utils.URIs.open_testdefault_cached()
hvmguest = Guest(testconn)
hvmguest.os.os_type = "hvm"
xenguest = Guest(testconn)
xenguest.os.os_type = "xen"

meter = virtinst.progress.make_meter(quiet=not utils.TESTCONFIG.debug)

if utils.TESTCONFIG.url_skip_libosinfo:
    os.environ["VIRTINST_TEST_SUITE_FORCE_LIBOSINFO"] = "0"
elif utils.TESTCONFIG.url_force_libosinfo:
    os.environ["VIRTINST_TEST_SUITE_FORCE_LIBOSINFO"] = "1"


def _sanitize_osdict_name(detectdistro):
    if detectdistro in ["none", "None", None]:
        return None
    return detectdistro


def _skipmsg(testdata):
    is_iso = testdata.url.lower().endswith(".iso")
    distname = testdata.name

    if utils.TESTCONFIG.url_iso_only and not is_iso:
        return "skipping non-iso test"
    elif utils.TESTCONFIG.url_only and is_iso:
        return "skipping non-url test"

    if not utils.TESTCONFIG.url_force_libosinfo:
        return
    if testdata.skip_libosinfo:
        return "force-libosinfo requested but test has skip_libosinfo set"
    if is_iso:
        return

    # If --force-libosinfo used, don't run tests that we know libosinfo
    # can't detect, non-treeinfo URLs basically
    if (
        "ubuntu" in distname
        or "debian" in distname
        or "mageia" in distname
        or "opensuse10" in distname
        or "opensuse11" in distname
        or "opensuse12" in distname
        or "opensuse13" in distname
        or "opensuseleap-42" in distname
        or "generic" in distname
        or testdata.url.startswith("ftp:/")
    ):
        return "skipping known busted libosinfo URL tests"


def _testGuest(testdata, guest):
    distname = testdata.name
    arch = testdata.arch
    url = testdata.url
    checkdistro = testdata.detectdistro

    guest.os.arch = arch
    guest.set_os_name("generic")
    if testdata.testshortcircuit:
        guest.set_os_name(checkdistro)

    msg = _skipmsg(testdata)
    if msg:
        raise pytest.skip(msg)

    installer = Installer(guest.conn, location=url)
    try:
        detected_distro = installer.detect_distro(guest)
    except Exception as e:
        msg = (
            "\nFailed in installer detect_distro():\n"
            "name   = %s\n"
            "url    = %s\n\n%s" % (distname, url, str(e))
        )
        raise type(e)(msg).with_traceback(sys.exc_info()[2]) from None

    # Make sure the stores are reporting correct distro name/variant
    if checkdistro != detected_distro:
        raise AssertionError(
            "Detected OS did not match expected values:\n"
            "found   = %s\n"
            "expect  = %s\n\n"
            "testname = %s\n"
            "url      = %s\n" % (detected_distro, checkdistro, distname, url)
        )

    if guest is xenguest:
        return

    # Do this only after the distro detection, since we actually need
    # to fetch files for that part
    treemedia = installer._treemedia  # pylint: disable=protected-access
    fetcher = treemedia._cached_fetcher  # pylint: disable=protected-access

    def fakeAcquireFile(filename, fullurl=None):
        log.debug("Fake acquiring filename=%s fullurl=%s", filename, fullurl)
        return filename

    fetcher.acquireFile = fakeAcquireFile

    # Fetch regular kernel
    kernel, initrd, kernelargs = treemedia.prepare(guest, meter, None)
    dummy = initrd
    if testdata.kernelregex and not re.match(testdata.kernelregex, kernel):
        raise AssertionError(
            "kernel=%s but testdata.kernelregex='%s'" % (kernel, testdata.kernelregex)
        )

    if testdata.kernelarg == "None":
        if bool(kernelargs):
            raise AssertionError(
                "kernelargs='%s' but testdata.kernelarg='%s'" % (kernelargs, testdata.kernelarg)
            )
    elif testdata.kernelarg:
        if testdata.kernelarg != str(kernelargs).split("=")[0]:
            raise AssertionError(
                "kernelargs='%s' but testdata.kernelarg='%s'" % (kernelargs, testdata.kernelarg)
            )


def test001BadURL():
    badurl = "http://aksdkakskdfa-idontexist.com/foo/tree"

    with pytest.raises(ValueError, match=".*maybe you mistyped.*"):
        installer = Installer(hvmguest.conn, location=badurl)
        installer.detect_distro(hvmguest)

    # Non-existent cdrom fails
    with pytest.raises(ValueError, match=".*non-existent path.*"):
        installer = Installer(hvmguest.conn, cdrom="/not/exist/foobar")
        assert installer.detect_distro(hvmguest) is None

    # Ensure existing but non-distro file doesn't error
    installer = Installer(hvmguest.conn, cdrom="/dev/null")
    assert installer.detect_distro(hvmguest) is None


def _make_tests():
    import configparser

    cfg = configparser.ConfigParser()
    cfg.read("tests/data/test_urls.ini")

    manualpath = "~/.config/virt-manager/test_urls_manual.ini"
    if os.path.exists(os.path.expanduser(manualpath)):
        cfg.read(os.path.expanduser(manualpath))

    distros = []
    for name in cfg.sections():
        vals = dict(cfg.items(name))
        url = vals["url"]

        if "distro" not in vals:
            print("url needs an explicit distro= value: %s" % url)
            sys.exit(1)
        d = _URLTestData(
            name,
            url,
            _sanitize_osdict_name(vals["distro"]),
            vals.get("testxen", "0") == "1",
            vals.get("testshortcircuit", "0") == "1",
            vals.get("kernelarg", None),
            vals.get("kernelregex", None),
            vals.get("skiplibosinfo", "0") == "1",
        )
        distros.append(d)

    return distros


@pytest.mark.parametrize(
    "testdata",
    _make_tests(),
    ids=lambda d: d.name,
)
def testURL(testdata):
    """
    Test that our URL detection logic works for grabbing kernels
    """
    _testGuest(testdata, hvmguest)
    if testdata.testxen:
        _testGuest(testdata, xenguest)



# encoding=utf-8
#
# Copyright (C) 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os.path

import pytest

from virtinst import Guest
from virtinst import NodeDevice
from virtinst import DeviceHostdev

from tests import utils

# Requires XML_SANITIZE to parse correctly, see bug 1184131
funky_chars_xml = """
<device>
  <name>L3B2616</name>
  <capability type='LENOVOá'/>
</device>
"""

DATADIR = utils.DATADIR + "/nodedev/"


def _nodeDevFromName(conn, devname):
    node = conn.nodeDeviceLookupByName(devname)
    xml = node.XMLDesc(0)
    return NodeDevice(conn, xml)


def _testNode2DeviceCompare(conn, nodename, devfile, nodedev=None):
    devfile = os.path.join(DATADIR, "devxml", devfile)
    if not nodedev:
        nodedev = _nodeDevFromName(conn, nodename)

    dev = DeviceHostdev(conn)
    dev.set_from_nodedev(nodedev)
    dev.set_defaults(Guest(conn))
    utils.diff_compare(dev.get_xml() + "\n", devfile)


def check_version(conn, version):
    # pylint: disable=protected-access
    if conn.support._check_version(version):
        return

    msg = f"Skipping check due to version < {version}"
    raise pytest.skip(msg)


def testFunkyChars():
    # Ensure parsing doesn't fail
    conn = utils.URIs.open_testdriver_cached()
    dev = NodeDevice(conn, funky_chars_xml)
    assert dev.name == "L3B2616"
    assert dev.device_type == "LENOVO"


def testNetDevice():
    conn = utils.URIs.open_testdriver_cached()
    devname = "net_00_1c_25_10_b1_e4"
    dev = _nodeDevFromName(conn, devname)
    assert dev.name == devname
    assert dev.parent == "pci_8086_1049"
    assert dev.device_type == "net"
    assert dev.interface == "eth0"


def testPCIDevice():
    conn = utils.URIs.open_testdriver_cached()
    nodename = "pci_8086_10fb"
    obj = _nodeDevFromName(conn, nodename)
    assert obj.is_pci_sriov() is True
    nodename = "pci_8086_2448"
    obj = _nodeDevFromName(conn, nodename)
    assert obj.is_pci_bridge() is True


def testUSBDevDevice():
    conn = utils.URIs.open_testdriver_cached()
    devname = "usb_device_781_5151_2004453082054CA1BEEE"
    dev = _nodeDevFromName(conn, devname)
    assert dev.vendor_name == "SanDisk Corp."
    assert dev.product_name == "Cruzer Micro 256/512MB Flash Drive"

    devname = "usb_device_1d6b_1_0000_00_1a_0"
    dev = _nodeDevFromName(conn, devname)
    assert dev.is_usb_linux_root_hub() is True


def testSCSIDevice():
    conn = utils.URIs.open_testdriver_cached()
    devname = "pci_8086_2829_scsi_host_scsi_device_lun0"
    dev = _nodeDevFromName(conn, devname)
    assert dev.host == "0"
    assert dev.bus == "0"
    assert dev.target == "0"


def testStorageDevice():
    conn = utils.URIs.open_testdriver_cached()
    devname = "storage_serial_SATA_WDC_WD1600AAJS__WD_WCAP95119685"
    dev = _nodeDevFromName(conn, devname)
    assert dev.block == "/dev/sda"
    assert dev.drive_type == "disk"
    assert dev.media_available is None

    devname = "storage_model_DVDRAM_GSA_U1200N"
    dev = _nodeDevFromName(conn, devname)
    assert dev.media_label == "Fedora12_media"
    assert dev.media_available == 1


def testSCSIBus():
    conn = utils.URIs.open_testdriver_cached()
    devname = "pci_8086_2829_scsi_host_1"
    dev = _nodeDevFromName(conn, devname)
    assert dev.host == "2"


def testDRMDevice():
    conn = utils.URIs.open_testdriver_cached()
    devname = "drm_renderD129"
    dev = _nodeDevFromName(conn, devname)
    assert dev.devnodes[0].path == "/dev/dri/renderD129"
    assert dev.devnodes[0].node_type == "dev"
    assert dev.devnodes[1].path == "/dev/dri/by-path/pci-0000:00:02.0-render"
    assert dev.devnodes[1].node_type == "link"
    assert dev.is_drm_render() is True
    assert dev.get_devnode("frob")


def testDASDMdev():
    conn = utils.URIs.open_testdriver_cached()
    check_version(conn, "10.4.0")
    devname = "mdev_8e37ee90_2b51_45e3_9b25_bf8283c03110"
    dev = _nodeDevFromName(conn, devname)
    assert dev.name == devname
    assert dev.parent == "css_0_0_0023"
    assert dev.device_type == "mdev"
    assert dev.type_id == "vfio_ccw-io"


def testAPQNMdev():
    conn = utils.URIs.open_testdriver_cached()
    check_version(conn, "10.4.0")
    devname = "mdev_11f92c9d_b0b0_4016_b306_a8071277f8b9"
    dev = _nodeDevFromName(conn, devname)
    assert dev.name == devname
    assert dev.parent == "ap_matrix"
    assert dev.device_type == "mdev"
    assert dev.type_id == "vfio_ap-passthrough"


def testPCIMdev():
    conn = utils.URIs.open_testdriver_cached()
    check_version(conn, "10.4.0")
    devname = "mdev_4b20d080_1b54_4048_85b3_a6a62d165c01"
    dev = _nodeDevFromName(conn, devname)
    assert dev.name == devname
    assert dev.parent == "pci_0000_06_00_0"
    assert dev.device_type == "mdev"
    assert dev.type_id == "nvidia-11"
    assert dev.get_mdev_uuid() == "4b20d080-1b54-4048-85b3-a6a62d165c01"


def testPCIMdevNewFormat():
    conn = utils.URIs.open_testdriver_cached()
    check_version(conn, "10.4.0")
    devname = "mdev_35ceae7f_eea5_4f28_b7f3_7b12a3e62d3c_0000_06_00_0"
    dev = _nodeDevFromName(conn, devname)
    assert dev.name == devname
    assert dev.parent == "pci_0000_06_00_0"
    assert dev.device_type == "mdev"
    assert dev.type_id == "nvidia-11"
    assert dev.get_mdev_uuid() == "35ceae7f-eea5-4f28-b7f3-7b12a3e62d3c"


# NodeDevice 2 Device XML tests


def testNodeDev2USB1():
    conn = utils.URIs.open_testdriver_cached()
    nodename = "usb_device_781_5151_2004453082054CA1BEEE"
    devfile = "usbdev1.xml"
    _testNode2DeviceCompare(conn, nodename, devfile)


def testNodeDev2USB2():
    conn = utils.URIs.open_testdriver_cached()
    nodename = "usb_device_1d6b_2_0000_00_1d_7"
    devfile = "usbdev2.xml"
    nodedev = _nodeDevFromName(conn, nodename)

    _testNode2DeviceCompare(conn, nodename, devfile, nodedev=nodedev)


def testNodeDev2PCI():
    conn = utils.URIs.open_testdriver_cached()
    nodename = "pci_1180_592"
    devfile = "pcidev.xml"
    _testNode2DeviceCompare(conn, nodename, devfile)


def testNodeDevFail():
    conn = utils.URIs.open_testdriver_cached()
    nodename = "usb_device_1d6b_1_0000_00_1d_1_if0"
    devfile = ""

    # This should exist, since usbbus is not a valid device to
    # pass to a guest.
    with pytest.raises(ValueError):
        _testNode2DeviceCompare(conn, nodename, devfile)



# Copyright (C) 2020 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import unittest.mock

import pytest

import virtinst

import tests.utils
import tests.urlfetcher_mock


@unittest.mock.patch.dict(os.environ, {"VIRTINST_TEST_SUITE_FORCE_LIBOSINFO": "0"})
def _test(mockurl, distro=None, initrd=None, kernel=None, xen=False, iso=False):
    # pylint: disable=protected-access

    conn = tests.utils.URIs.open_testdefault_cached()
    guest = virtinst.Guest(conn)
    guest.os.os_type = xen and "xen" or "hvm"

    url = tests.urlfetcher_mock.make_mock_input_url(mockurl)
    installer = virtinst.Installer(guest.conn, location=url)
    treemedia = installer._treemedia
    treemedia._get_fetcher(guest, None)
    if iso:
        treemedia._cached_fetcher._is_iso = True

    detected_distro = installer.detect_distro(guest)
    assert (distro or "") in (detected_distro or "")

    # Fetch regular kernel
    treecache = treemedia._cached_data
    kernels = [p[0] for p in treecache.kernel_pairs]
    initrds = [p[1] for p in treecache.kernel_pairs]
    if initrd:
        assert any([i.endswith(initrd) for i in initrds])
    if kernel:
        assert any([k.endswith(kernel) for k in kernels])


def test_debian():
    _test("debian/buster/main/installer-amd64", distro="debian10", kernel="linux")
    _test(
        "debian/buster/main/installer-amd64",
        distro="debian10",
        xen=True,
        kernel="netboot/xen/vmlinuz",
    )
    _test("debian/buster/main/installer-s390x", distro="debian10", kernel="kernel.debian")
    _test("debian/buster/main/installer-ppc64el", distro="debian10", kernel="vmlinux")
    _test("debian/buster/main/installer-arm64", distro="debian10")
    _test("debian/daily-images/amd64", distro="debiantesting")

    _test("debian/debian-8.10.0-amd64-netinst.iso", kernel="linux")
    _test(
        "debian/debian-8.10.0-amd64-netinst.iso",
        distro="debian8",
        iso=True,
        kernel="install.amd/vmlinuz",
    )
    _test("debian/debian-8.10.0-s390x-netinst.iso", iso=True, kernel="linux_vm")
    _test("debian/debian-8.10.0-ppc64el-netinst.iso", iso=True, kernel="vmlinux")
    _test("debian/debian-8.10.0-i386-netinst.iso", iso=True, kernel="install.386/vmlinuz")
    _test(
        "debian/debian-8.10.0-arm64-netinst.iso",
        iso=True,
        kernel="install.a64/vmlinuz",
    )
    # Bad arch triggers a fallback path
    _test(
        "debian/debian-8.10.0-badarch-netinst.iso",
        iso=True,
        kernel="install/vmlinuz",
    )
    # Fails to detect treearch, hits certain paths
    _test("debian/debian-8.10.0-badarch-netinst.iso", kernel="linux")


def test_ubuntu():
    _test("ubuntu/bionic/main/installer-amd64", "ubuntu18.04")
    _test("ubuntu/focal/main/installer-amd64", "ubuntu20.04")

    _test("ubuntu/ubuntu-17.10-amd64.iso", iso=True, kernel="install/vmlinuz")
    _test("ubuntu/ubuntu-17.10-s390x.iso", iso=True, kernel="boot/kernel.ubuntu")


def test_fedora():
    _test("fedora/30", "fedora30")
    _test("fedora/rawhide", "fedora-unknown")
    # Fake fedora version 99 to hit certain code paths
    _test("fedora/99", "fedora-unknown")


def test_rhel():
    _test("rhel/7.6", "rhel7.6")
    # Fake rhel 7.20 to hit certain code paths
    _test("rhel/7.20", "rhel7.")


def test_centos():
    _test("centos/6.10", "centos6.10")
    _test("centos/sl7", "centos7.0")


def test_opensuse():
    _test("opensuse/tumbleweed", "opensusetumbleweed")
    _test("opensuse/10.3", "opensuse10.3")
    _test("opensuse/11.4", "opensuse11.4")
    _test("opensuse/12.3", "opensuse12.3")
    _test("opensuse/13.2", "opensuse13.2", xen=True, initrd="initrd-xen")
    # Specifically use trailing slash to hit url scraping code path
    _test("opensuse/42.3/", "opensuse42.3")
    # Fake version to trigger particular path in urldetect.py
    _test("opensuse/15.9", "opensuse15")
    # Has a bad version number which isn't in osinfo-db
    _test("opensuse/badversion/", None)


def test_suse():
    _test("suse/SLES-10-SP4-DVD-x86_64-GM-DVD1.iso", "sles10sp4")
    _test("suse/SLES-11-SP4-DVD-s390x-GM-DVD1.iso", "sles11sp4", kernel="vmrdr.ikr")
    _test("suse/SLES-11-SP4-DVD-ppc64-GM-DVD1.iso", "sles11sp4", kernel="linux64")


def test_mageia():
    _test("mageia/5", "mageia5", initrd="all.rdz")
    _test("mageia/8", initrd="all.rdz")


def test_misc():
    _test("generic")

    with pytest.raises(ValueError) as e:
        _test("empty")
    assert "installable distribution" in str(e.value)
    assert "mistyped" in str(e.value)



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from . import lib


#######################################
# UI tests for the preferences dialog #
#######################################


def testPrefsAll(app):
    app.root.find("Edit", "menu").click()
    app.root.find("Preferences", "menu item").click()

    win = app.find_window("Preferences")
    generaltab = win.find("general-tab")
    pollingtab = win.find("polling-tab")
    newvmtab = win.find("newvm-tab")
    consoletab = win.find("console-tab")
    feedbacktab = win.find("feedback-tab")

    feedbacktab.check_not_onscreen()
    tab = generaltab
    tab.check_onscreen()
    tab.find_fuzzy("Enable system tray", "check").click()
    tab.find_fuzzy("Enable XML").click()
    tab.find_fuzzy("libguestfs VM").click()

    win.find("Polling", "page tab").click()
    tab = pollingtab
    tab.check_onscreen()
    tab.find("Poll CPU", "check box").click()
    tab.find("Poll Disk", "check box").click()
    tab.find("Poll Memory", "check box").click()
    tab.find("Poll Network", "check box").click()
    period = tab.find_fuzzy("cpu-poll", "spin button")
    period.click()
    period.set_text("5")

    win.find("New VM", "page tab").click()
    tab = newvmtab
    tab.check_onscreen()
    tab.combo_select("CPU default:", "host-passthrough")
    tab.combo_select("Storage format:", "Raw")
    tab.combo_select("Graphics type", "VNC")
    tab.combo_select("x86 Firmware", "UEFI")

    win.find("Console", "page tab").click()
    tab = consoletab
    tab.check_onscreen()
    tab.combo_select("SPICE USB", "Manual redirect")
    tab.combo_select("Resize guest", "On")
    tab.combo_select("Graphical console scaling", "Never")
    tab.find("Console autoconnect", "check box").click()

    tab.find("Change...", "push button").click()
    keyframe = app.find_window("Configure grab")

    # On certain environments pressing "Alt_L" and
    # clicking a window starts window drag operation.
    # Work around by pushing both Control and Alt.
    app.rawinput.holdKey("Control_L")
    app.rawinput.holdKey("Alt_L")
    app.rawinput.holdKey("Z")

    # Test releasekey handler
    app.rawinput.releaseKey("Z")
    app.rawinput.holdKey("Z")
    try:
        keyframe.find_fuzzy("OK", "push button").click()
    finally:
        app.rawinput.releaseKey("Z")
        app.rawinput.releaseKey("Alt_L")
        app.rawinput.releaseKey("Control_L")

    win.find("Feedback", "page tab").click()
    tab = feedbacktab
    tab.check_onscreen()
    tab.find("Force Poweroff", "check box").click()
    tab.find("Poweroff/Reboot", "check box").click()
    tab.find("Pause", "check box").click()
    tab.find("Device removal", "check box").click()
    tab.find("Unapplied changes", "check box").click()
    tab.find("Deleting storage", "check box").click()

    win.find("General", "page tab").click()
    win.find_fuzzy("Enable system tray", "check").click()

    win.find_fuzzy("Close", "push button").click()
    lib.utils.check(lambda: win.visible is False)


def testPrefsXMLEditor(app):
    managerwin = app.topwin
    detailswin = app.manager_open_details("test-clone-simple")
    finish = detailswin.find("config-apply")
    xmleditor = detailswin.find("XML editor")

    detailswin.find("XML", "page tab").click()
    warnlabel = detailswin.find_fuzzy("XML editing is disabled")
    lib.utils.check(lambda: warnlabel.visible)
    origtext = xmleditor.text
    xmleditor.typeText("1234abcd")
    lib.utils.check(lambda: xmleditor.text == origtext)

    managerwin.grab_focus()
    managerwin.find("Edit", "menu").click()
    managerwin.find("Preferences", "menu item").click()
    prefswin = app.find_window("Preferences")
    prefswin.find_fuzzy("Enable XML").click()
    prefswin.find_fuzzy("Close", "push button").click()
    lib.utils.check(lambda: prefswin.visible is False)

    managerwin.window_close()
    detailswin.grab_focus()
    newtext = xmleditor.text.replace(">", "><title>FOOTITLE</title>", 1)
    xmleditor.set_text(newtext)
    finish.click()
    detailswin.find("Details", "page tab").click()
    lib.utils.check(lambda: detailswin.find("Title:", "text").text == "FOOTITLE")


def testPrefsKeyfile(app):
    """
    Preload some keyfile settings and verify they work as expected
    """
    app.open(use_uri=False, keyfile="defaultconn.ini")
    managerwin = app.topwin

    # test:///default should be connected
    managerwin.find("test default", "table cell")
    managerwin.find("foo - Not Connected", "table cell")



#
# Common code for all guests
#
# Copyright 2006-2009, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

from . import unattended
from . import volumeupload
from .installertreemedia import InstallerTreeMedia
from .installerinject import perform_cdrom_injections
from ..domain import DomainOs
from ..devices import DeviceDisk
from ..guest import Guest
from ..osdict import OSDB
from ..logger import log
from .. import progress
from .. import xmlutil


def _make_testsuite_path(path):
    return os.path.join("/VIRTINST-TESTSUITE", os.path.basename(path).split("-", 2)[-1])


class Installer:
    """
    Class for kicking off VM installs. The VM is set up separately in a Guest
    instance. This class tracks the install media/bootdev choice, alters the
    Guest XML, boots it for the install, then saves the post install XML
    config. The Guest is passed in via start_install, only install media
    selection is done at __init__ time

    :param cdrom: Path to a cdrom device or iso. Maps to virt-install --cdrom
    :param location: An install tree URI, local directory, or ISO/CDROM path.
        Largely handled by installtreemedia helper class. Maps to virt-install
        --location
    :param location_kernel: URL pointing to a kernel to fetch, or a relative
        path to indicate where the kernel is stored in location
    :param location_initrd: location_kernel, but pointing to an initrd
    :param install_kernel: Kernel to install off of
    :param install_initrd: Initrd to install off of
    :param install_kernel_args: Kernel args <cmdline> to use. This overwrites
        whatever the installer might request, unlike extra_args which will
        append arguments.
    :param no_install: If True, this installer specifically does not have
        an install phase. We are just using it to create the initial XML.
    """

    def __init__(
        self,
        conn,
        cdrom=None,
        location=None,
        install_bootdev=None,
        location_kernel=None,
        location_initrd=None,
        install_kernel=None,
        install_initrd=None,
        install_kernel_args=None,
        no_install=None,
        is_reinstall=False,
    ):
        self.conn = conn

        # Entry point for virt-manager 'Customize' wizard to change autostart
        self.autostart = False

        self._install_cdrom_device_added = False
        self._unattended_install_cdrom_device = None
        self._tmpfiles = []
        self._tmpvols = []
        self._defaults_are_set = False
        self._unattended_data = None
        self._cloudinit_data = None

        self._install_bootdev = install_bootdev
        self._no_install = no_install
        self._is_reinstall = is_reinstall
        self._pre_reinstall_xml = None

        self._treemedia = None
        self._treemedia_bootconfig = None
        self._cdrom = None
        if cdrom:
            cdrom = InstallerTreeMedia.validate_path(self.conn, cdrom)
            self._cdrom = cdrom
            self._install_bootdev = "cdrom"
        elif location or location_kernel or location_initrd or install_kernel or install_initrd:
            self._treemedia = InstallerTreeMedia(
                self.conn,
                location,
                location_kernel,
                location_initrd,
                install_kernel,
                install_initrd,
                install_kernel_args,
            )

    ##################
    # Static helpers #
    ##################

    @staticmethod
    def cleanup_created_disks(guest, meter):
        """
        Remove any disks we created as part of the install. Only ever
        called by clients.
        """
        clean_disks = [d for d in guest.devices.disk if d.storage_was_created]

        for disk in clean_disks:
            path = disk.get_source_path()
            log.debug("Removing created disk path=%s vol_object=%s", path, disk.get_vol_object())
            name = os.path.basename(path)

            try:
                meter.start(_("Removing disk '%s'") % name, None)

                if disk.get_vol_object():
                    disk.get_vol_object().delete()
                else:  # pragma: no cover
                    # This case technically shouldn't happen here, but
                    # it's here in case future assumptions change
                    os.unlink(path)

                meter.end()
            except Exception as e:  # pragma: no cover
                log.debug("Failed to remove disk '%s'", name, exc_info=True)
                log.error("Failed to remove disk '%s': %s", name, e)

    ###################
    # Private helpers #
    ###################

    def _make_cdrom_device(self, path):
        dev = DeviceDisk(self.conn)
        dev.device = dev.DEVICE_CDROM
        dev.set_source_path(path)
        dev.sync_path_props()
        dev.validate()
        return dev

    def _cdrom_path(self):
        if self._treemedia:
            return self._treemedia.cdrom_path()
        return self._cdrom

    def _add_install_cdrom_device(self, guest):
        if self._install_cdrom_device_added:
            return  # pragma: no cover
        if not bool(self._cdrom_path()):
            return

        self._install_cdrom_device_added = True

        if self._is_reinstall:
            cdroms = [d for d in guest.devices.disk if d.is_cdrom()]
            if cdroms:
                dev = cdroms[0]
                dev.set_source_path(self._cdrom_path())
                return

        dev = self._make_cdrom_device(self._cdrom_path())
        if self._is_reinstall:
            dev.set_defaults(guest)

        # Insert the CDROM before any other CDROM, so boot=cdrom picks
        # it as the priority
        for idx, disk in enumerate(guest.devices.disk):
            if disk.is_cdrom():
                guest.devices.add_child(dev, idx=idx)
                return
        guest.add_device(dev)

    def _remove_install_cdrom_media(self, guest):
        if not self._install_cdrom_device_added:
            return
        if guest.osinfo.is_windows():
            # Keep media attached for windows which has a multi stage install
            return
        for disk in guest.devices.disk:
            if disk.is_cdrom() and disk.get_source_path() == self._cdrom_path():
                disk.set_source_path(None)
                disk.sync_path_props()
                break

    def _add_unattended_install_cdrom_device(self, guest, location):
        if self._unattended_install_cdrom_device:
            return  # pragma: no cover

        dev = self._make_cdrom_device(location)
        dev.set_defaults(guest)
        self._unattended_install_cdrom_device = dev.target
        guest.add_device(dev)

        # The CDROM we just added might be a SCSI one. If that's the case,
        # libvirt will automatically add the necessary controller for us, but
        # depending on the architecture its choice of model might not be
        # usable by the guest OS. Try to add a virtio-scsi controller, if
        # supported, for the best chance of the CDROM actually being picked up
        guest.add_virtioscsi_controller()

        if self.conn.in_testsuite():
            # Hack to set just the XML path differently for the test suite.
            # Setting this via regular 'path' will error that it doesn't exist
            dev.source.file = _make_testsuite_path(location)

    def _remove_unattended_install_cdrom_device(self, guest):
        if not self._unattended_install_cdrom_device:
            return

        disk = [d for d in guest.devices.disk if d.target == self._unattended_install_cdrom_device][
            0
        ]
        disk.set_source_path(None)
        disk.sync_path_props()

    def _build_boot_order(self, guest, bootdev):
        bootorder = [bootdev]

        # If guest has an attached disk, always have 'hd' in the boot
        # list, so disks are marked as bootable/installable (needed for
        # windows virtio installs, and booting local disk from PXE)
        for disk in guest.devices.disk:
            if disk.device == disk.DEVICE_DISK:
                bootdev = "hd"
                if bootdev not in bootorder:
                    bootorder.append(bootdev)
                break
        return bootorder

    def _can_set_guest_bootorder(self, guest):
        return (
            not guest.os.is_container()
            and not guest.os.kernel
            and not any([d.boot.order for d in guest.devices.get_all()])
        )

    def _alter_treemedia_bootconfig(self, guest):
        if not self._treemedia:
            return

        kernel, initrd, kernel_args = self._treemedia_bootconfig
        if kernel:
            guest.os.kernel = self.conn.in_testsuite() and _make_testsuite_path(kernel) or kernel
        if initrd:
            guest.os.initrd = self.conn.in_testsuite() and _make_testsuite_path(initrd) or initrd
        if kernel_args:
            guest.os.kernel_args = kernel_args

    def _alter_bootconfig(self, guest):
        """
        Generate the portion of the guest xml that determines boot devices
        and parameters. (typically the <os></os> block)

        :param guest: Guest instance we are installing
        """
        guest.on_reboot = "destroy"
        self._alter_treemedia_bootconfig(guest)

        bootdev = self._install_bootdev
        if bootdev and self._can_set_guest_bootorder(guest):
            guest.os.bootorder = self._build_boot_order(guest, bootdev)
        else:
            guest.os.bootorder = []

    def _alter_install_resources(self, guest, meter):
        """
        Sets the appropriate amount of ram needed when performing a "network"
        based installation

        :param guest: Guest instance we are installing
        """
        if not self._treemedia:
            return
        if not self._treemedia.requires_internet(guest, meter):
            return

        ram = guest.osinfo.get_network_install_required_ram(guest)
        ram = (ram or 0) // 1024
        if ram > guest.currentMemory:
            msg = _(
                "Overriding memory to %(number)s MiB needed for %(osname)s network install."
            ) % {"number": ram // 1024, "osname": guest.osinfo.name}
            log.warning(msg)
            guest.currentMemory = ram

    ################
    # Internal API #
    ################

    def _should_upload_media(self, guest):
        """
        Return True if we should upload media to the connection scratchdir.
        This doesn't consider if there is any media to upload, just whether
        we _should_ upload if there _is_ media.
        """
        scratchdir = InstallerTreeMedia.make_scratchdir(guest)
        system_scratchdir = InstallerTreeMedia.get_system_scratchdir(guest)

        if self.conn.is_remote():
            return True
        if self.conn.is_unprivileged():
            return False
        if scratchdir == system_scratchdir:
            return False  # pragma: no cover
        return True

    def _upload_media(self, guest, meter, paths):
        system_scratchdir = InstallerTreeMedia.get_system_scratchdir(guest)

        if not self._should_upload_media(guest) and not xmlutil.in_testsuite():
            # We have access to system scratchdir, don't jump through hoops
            log.debug(
                "Have access to preferred scratchdir so nothing to upload"
            )  # pragma: no cover
            return paths  # pragma: no cover

        if not guest.conn.support_remote_url_install():
            # Needed for the test_urls suite
            log.debug("Media upload not supported")  # pragma: no cover
            return paths  # pragma: no cover

        newpaths, tmpvols = volumeupload.upload_paths(guest.conn, system_scratchdir, meter, paths)
        self._tmpvols += tmpvols
        return newpaths

    def _prepare_unattended_data(self, guest, meter, scripts):
        scratchdir = InstallerTreeMedia.make_scratchdir(guest)

        injections = []
        for script in scripts:
            expected_filename = script.get_expected_filename()
            unattended_cmdline = script.generate_cmdline()
            log.debug("Generated unattended cmdline: %s", unattended_cmdline)

            scriptpath = script.write()
            self._tmpfiles.append(scriptpath)
            injections.append((scriptpath, expected_filename))

        drivers_location = guest.osinfo.get_pre_installable_drivers_location(guest.os.arch)
        drivers = unattended.download_drivers(drivers_location, scratchdir, meter)
        injections.extend(drivers)
        self._tmpfiles.extend([driverpair[0] for driverpair in drivers])

        iso = perform_cdrom_injections(injections, scratchdir)
        self._tmpfiles.append(iso)
        iso = self._upload_media(guest, meter, [iso])[0]
        self._add_unattended_install_cdrom_device(guest, iso)

    def _prepare_unattended_scripts(self, guest, meter):
        url = None
        os_tree = None
        if self._treemedia:
            if self._treemedia.is_network_url():
                url = self.location
            os_media = self._treemedia.get_os_media(guest, meter)
            os_tree = self._treemedia.get_os_tree(guest, meter)
            injection_method = "initrd"
        else:
            if not guest.osinfo.is_windows():
                log.warning(
                    "Attempting unattended method=cdrom injection "
                    "for a non-windows OS. If this doesn't work, try "
                    "passing install media to --location"
                )
            osguess = OSDB.guess_os_by_iso(self.cdrom)
            os_media = osguess[1] if osguess else None
            injection_method = "cdrom"

        return unattended.prepare_install_scripts(
            guest, self._unattended_data, url, os_media, os_tree, injection_method
        )

    def _prepare_treemedia(self, guest, meter, unattended_scripts):
        kernel, initrd, kernel_args = self._treemedia.prepare(guest, meter, unattended_scripts)

        paths = [kernel, initrd]
        kernel, initrd = self._upload_media(guest, meter, paths)
        self._treemedia_bootconfig = (kernel, initrd, kernel_args)

    def _prepare_cloudinit(self, guest, meter):
        scratchdir = InstallerTreeMedia.make_scratchdir(guest)
        filepairs = self._cloudinit_data.create_files(scratchdir)
        for filepair in filepairs:
            self._tmpfiles.append(filepair[0])

        iso = perform_cdrom_injections(filepairs, scratchdir, cloudinit=True)
        self._tmpfiles.append(iso)
        iso = self._upload_media(guest, meter, [iso])[0]
        self._add_unattended_install_cdrom_device(guest, iso)

    def _prepare(self, guest, meter):
        if self._is_reinstall:
            self._pre_reinstall_xml = guest.get_xml()

        unattended_scripts = None
        if self._unattended_data:
            unattended_scripts = self._prepare_unattended_scripts(guest, meter)

        if self._treemedia:
            self._prepare_treemedia(guest, meter, unattended_scripts)

        elif unattended_scripts:
            self._prepare_unattended_data(guest, meter, unattended_scripts)

        if self._cloudinit_data:
            self._prepare_cloudinit(guest, meter)

    def _cleanup(self, guest):
        if self._treemedia:
            self._treemedia.cleanup(guest)

        for vol in self._tmpvols:
            log.debug("Removing volume '%s'", vol.name())
            vol.delete(0)
        self._tmpvols = []

        for f in self._tmpfiles:
            log.debug("Removing %s", str(f))
            os.unlink(f)
        self._tmpfiles = []

    def _get_postinstall_bootdev(self, guest):
        if self.cdrom and self._no_install:
            return DomainOs.BOOT_DEVICE_CDROM

        if self._install_bootdev:
            if any([d for d in guest.devices.disk if d.device == d.DEVICE_DISK]):
                return DomainOs.BOOT_DEVICE_HARDDISK
            return self._install_bootdev

        device = guest.devices.disk and guest.devices.disk[0].device or None
        if device == DeviceDisk.DEVICE_DISK:
            return DomainOs.BOOT_DEVICE_HARDDISK
        elif device == DeviceDisk.DEVICE_CDROM:
            return DomainOs.BOOT_DEVICE_CDROM
        elif device == DeviceDisk.DEVICE_FLOPPY:
            return DomainOs.BOOT_DEVICE_FLOPPY
        return DomainOs.BOOT_DEVICE_HARDDISK

    ##############
    # Public API #
    ##############

    @property
    def location(self):
        if self._treemedia:
            return self._treemedia.location

    @property
    def cdrom(self):
        return self._cdrom

    def set_initrd_injections(self, initrd_injections):
        if not self._treemedia:
            raise RuntimeError("Install method does not support initrd injections.")
        self._treemedia.set_initrd_injections(initrd_injections)

    def set_extra_args(self, extra_args):
        if not self._treemedia:
            raise RuntimeError(
                "Kernel arguments are only supported with location or kernel installs."
            )
        self._treemedia.set_extra_args(extra_args)

    def set_install_defaults(self, guest):
        """
        Allow API users to set defaults ahead of time if they want it.
        Used by vmmDomainVirtinst so the 'Customize before install' dialog
        shows accurate values.

        If the user doesn't explicitly call this, it will be called by
        start_install()
        """
        if self._defaults_are_set:
            return
        if self._is_reinstall:
            self._pre_reinstall_xml = guest.get_xml()

        self._add_install_cdrom_device(guest)

        if not guest.os.bootorder and self._can_set_guest_bootorder(guest):
            bootdev = self._get_postinstall_bootdev(guest)
            guest.os.bootorder = self._build_boot_order(guest, bootdev)

        if not self._is_reinstall:
            guest.set_defaults(None)
        self._defaults_are_set = True

    def get_search_paths(self, guest):
        """
        Return a list of paths that the hypervisor will need search access
        for to perform this install.
        """
        search_paths = []
        if (
            self._treemedia or self._cloudinit_data or self._unattended_data
        ) and not self._should_upload_media(guest):
            search_paths.append(InstallerTreeMedia.make_scratchdir(guest))
        if self._cdrom_path():
            search_paths.append(self._cdrom_path())
        return search_paths

    def requires_postboot_xml_changes(self):
        """
        Return True if the install operation has 2 XML phases, and
        requires a hard VM poweroff to ensure the second stage is
        used for subsequent boots.
        """
        if self.has_cloudinit() or self.has_unattended():
            return True
        if self._no_install:
            return False
        return bool(self._cdrom or self._install_bootdev or self._treemedia)

    def options_specified(self):
        """
        Return True if some explicit install option was actually passed in
        """
        if self._no_install:
            return True
        return self.requires_postboot_xml_changes()

    def detect_distro(self, guest):
        """
        Attempt to detect the distro for the Installer's 'location'. If
        an error is encountered in the detection process (or if detection
        is not relevant for the Installer type), None is returned.

        :returns: distro variant string, or None
        """
        ret = None
        if self._treemedia:
            ret = self._treemedia.detect_distro(guest)
        elif self.cdrom:
            if guest.conn.is_remote():
                log.debug("Can't detect distro for cdrom remote connection.")
            else:
                osguess = OSDB.guess_os_by_iso(self.cdrom)
                if osguess:
                    ret = osguess[0]
        else:
            log.debug("No media for distro detection.")

        log.debug("installer.detect_distro returned=%s", ret)
        return ret

    def set_unattended_data(self, unattended_data):
        self._unattended_data = unattended_data

    def set_cloudinit_data(self, cloudinit_data):
        self._cloudinit_data = cloudinit_data

    def has_cloudinit(self):
        return bool(self._cloudinit_data)

    def has_unattended(self):
        return bool(self._unattended_data)

    def get_generated_password(self):
        if self._cloudinit_data:
            return self._cloudinit_data.get_password_if_generated()

    ##########################
    # guest install handling #
    ##########################

    def _build_postboot_xml(self, guest_ro, final_xml, meter):
        initial_guest = Guest(self.conn, parsexml=final_xml)
        self._alter_bootconfig(initial_guest)
        self._alter_install_resources(initial_guest, meter)
        if self.has_cloudinit():
            initial_guest.set_smbios_serial_cloudinit()

            # When shim in the guest sees unpopulated EFI NVRAM, like when
            # we create a new UEFI VM, it invokes fallback.efi to populate
            # initial NVRAM boot entries. When the guest also has a TPM device,
            # shim will do a one time VM reset. This reset throws off the
            # reboot detection that is central to virt-install's install
            # process.
            #
            # The main install case that this will usually be relevant is
            # the combo of UEFI and --cloud-init. The latter usually implies
            # use of a distro cloud image, which will be using shim, and the
            # --cloud-init process requires a multi stage install compared
            # to just a plain import install.
            #
            # For that case, we disable the default TPM device for the first
            # boot.
            if guest_ro.have_default_tpm and guest_ro.is_uefi() and len(initial_guest.devices.tpm):
                log.debug(
                    "combo of default TPM, UEFI, and cloudinit is "
                    "used. assuming this VM is using a linux distro "
                    "cloud image with shim in the boot path. disabling "
                    "TPM for the first boot"
                )
                initial_guest.remove_device(initial_guest.devices.tpm[0])

        final_guest = Guest(self.conn, parsexml=final_xml)
        self._remove_install_cdrom_media(final_guest)
        self._remove_unattended_install_cdrom_device(final_guest)

        return initial_guest.get_xml(), final_guest.get_xml()

    def _build_xml(self, guest, meter):
        initial_xml = None
        final_xml = guest.get_xml()
        if self.requires_postboot_xml_changes():
            initial_xml, final_xml = self._build_postboot_xml(guest, final_xml, meter)
        final_xml = self._pre_reinstall_xml or final_xml

        log.debug(
            "Generated initial_xml: %s", (initial_xml and ("\n" + initial_xml) or "None required")
        )
        log.debug("Generated final_xml: \n%s", final_xml)

        return initial_xml, final_xml

    def _manual_transient_create(self, initial_xml, final_xml, needs_boot):
        """
        For hypervisors (like vz) that don't implement createXML,
        we need to define+start, and undefine on start failure
        """
        domain = self.conn.defineXML(initial_xml or final_xml)
        if not needs_boot:
            return domain

        # Handle undefining the VM if the initial startup fails
        try:
            domain.create()
        except Exception:  # pragma: no cover
            try:
                domain.undefine()
            except Exception:
                pass
            raise

        if initial_xml and initial_xml != final_xml:
            domain = self.conn.defineXML(final_xml)
        return domain

    def _create_guest(self, guest, meter, initial_xml, final_xml, doboot, transient):
        """
        Actually do the XML logging, guest defining/creating

        :param doboot: Boot guest even if it has no install phase
        """
        meter_label = _("Creating domain...")
        meter = progress.ensure_meter(meter)
        meter.start(meter_label, None)
        needs_boot = doboot or bool(initial_xml)

        if guest.type == "vz" and not self._is_reinstall:
            if transient:
                raise RuntimeError(_("Domain type 'vz' doesn't support transient installs."))
            domain = self._manual_transient_create(initial_xml, final_xml, needs_boot)

        else:
            if transient or needs_boot:
                domain = self.conn.createXML(initial_xml or final_xml, 0)
            if not transient:
                domain = self.conn.defineXML(final_xml)

        try:
            log.debug("XML fetched from libvirt object:\n%s", domain.XMLDesc(0))
        except Exception as e:  # pragma: no cover
            log.debug("Error fetching XML from libvirt object: %s", e)
        meter.end()
        return domain

    def _flag_autostart(self, domain):
        """
        Set the autostart flag for domain if the user requested it
        """
        try:
            domain.setAutostart(True)
        except Exception as e:  # pragma: no cover
            if not self.conn.support.is_error_nosupport(e):
                raise
            log.warning(
                "Could not set autostart flag: libvirt connection does not support autostart."
            )

    ######################
    # Public install API #
    ######################

    def start_install(
        self, user_guest, meter=None, dry=False, return_xml=False, doboot=True, transient=False
    ):
        """
        Begin the guest install. Will add install media to the guest config,
        launch it, then redefine the XML with the postinstall config.

        :param return_xml: Don't create the guest, just return generated XML
        """
        if not self._is_reinstall and not return_xml:
            Guest.validate_name(self.conn, user_guest.name)
        self.set_install_defaults(user_guest)
        disks = user_guest.devices.disk[:]

        # All installer XML alterations are made on this guest instance,
        # so the user_guest instance is left intact
        guest = Guest(self.conn, parsexml=user_guest.get_xml())
        guest.have_default_tpm = user_guest.have_default_tpm

        try:
            self._prepare(guest, meter)

            if not dry and not self._is_reinstall:
                for dev in disks:
                    dev.build_storage(meter)

            initial_xml, final_xml = self._build_xml(guest, meter)
            if dry or return_xml:
                return (initial_xml, final_xml)

            domain = self._create_guest(guest, meter, initial_xml, final_xml, doboot, transient)

            if self.autostart:
                self._flag_autostart(domain)
            return domain
        finally:
            self._cleanup(guest)



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import tests.utils
from . import lib


###################################
# UI tests for the migrate dialog #
###################################


def _open_migrate(app, vmname):
    app.manager_vm_action(vmname, migrate=True)
    return app.find_window("Migrate the virtual machine")


def testMigrateQemu(app):
    # Use fake qemu connections
    app.uri = tests.utils.URIs.kvm_x86
    newuri = tests.utils.URIs.test_default + ",fakeuri=qemu+tcp://fakehost/system"
    app.manager_createconn(newuri)

    # Run default migrate
    mig = _open_migrate(app, "test-many-devices")
    mig.find("Migrate", "push button").click()
    app.click_alert_button("the.connection.driver:.virDomainMigrate", "Close")
    mig.find("Cancel", "push button").click()
    lib.utils.check(lambda: not mig.showing)

    # Run with deselected URI
    mig = _open_migrate(app, "test-many-devices")
    mig.find("address-check").click()
    label = mig.find("Let libvirt decide")
    label.check_onscreen()
    mig.find("Migrate", "push button").click()
    app.click_alert_button("the.connection.driver:.virDomainMigrate", "Close")
    mig.find("Cancel", "push button").click()
    lib.utils.check(lambda: not mig.showing)

    # Run with tunnelled and other options
    mig = _open_migrate(app, "test-many-devices")
    mig.combo_select("Mode:", "Tunnelled")
    mig.find("Advanced", "toggle button").click_expander()
    mig.find("Allow unsafe:", "check box").click()
    mig.find("Temporary", "check box").click()

    mig.find("Migrate", "push button").click()
    app.click_alert_button("p2p migration", "Close")
    mig.find("Cancel", "push button").click()
    lib.utils.check(lambda: not mig.showing)


def testMigrateXen(app):
    # Use fake xen connections
    app.uri = tests.utils.URIs.test_full + ",fakeuri=xen:///"

    fakeremotexen = tests.utils.URIs.test_default + ",fakeuri=xen+tcp://fakehost/"
    app.manager_createconn(fakeremotexen)

    # Run default migrate
    mig = _open_migrate(app, "test-many-devices")
    mig.find("Migrate", "push button").click()
    app.click_alert_button("the.connection.driver:.virDomainMigrate", "Close")
    mig.find("Cancel", "push button").click()
    lib.utils.check(lambda: not mig.showing)


def testMigrateMock(app):
    """
    Trigger the mock migration testing we have to emulate success
    """
    # Add an additional connection
    app.manager_createconn("test:///default")

    # Run it and check some values
    mig = _open_migrate(app, "test-many-devices")
    mig.find("address-text").set_text("TESTSUITE-FAKE")

    mig.find("Migrate", "push button").click()
    progwin = app.find_window("Migrating VM")
    # Attempt cancel which will fail, then find the error message
    progwin.find("Cancel", "push button").click()
    progwin.find("Error cancelling migrate job")
    lib.utils.check(lambda: not progwin.showing, timeout=5)
    lib.utils.check(lambda: not mig.showing)


def testMigrateConnMismatch(app):
    # Add a possible target but disconnect it
    app.uri = tests.utils.URIs.test_default
    manager = app.topwin
    manager.window_maximize()
    manager.click()
    app.manager_conn_disconnect("test default")

    # Add a mismatched hv connection
    fakexen = tests.utils.URIs.test_empty + ",fakeuri=xen:///"
    app.manager_createconn(fakexen)

    # Open dialog and confirm no conns are available
    app.manager_createconn(tests.utils.URIs.test_full)
    mig = _open_migrate(app, "test-many-devices")
    mig.find("conn-combo").find("No usable", "menu item")

    # Test explicit dialog 'delete'
    mig.window_close()

    # Ensure disconnecting will close the dialog
    manager.click_title()
    mig = _open_migrate(app, "test-many-devices")
    app.manager_conn_disconnect("test testdriver.xml")
    lib.utils.check(lambda: not mig.showing)


def testMigrateXMLEditor(app):
    app.open(xmleditor_enabled=True)
    manager = app.topwin

    # Add an additional connection
    app.manager_createconn("test:///default")

    # Run it and check some values
    vmname = "test-many-devices"
    win = _open_migrate(app, vmname)
    win.find("address-text").set_text("TESTSUITE-FAKE")

    # Create a new obj with XML edited name, verify it worked
    newname = "aafroofroo"
    win.find("XML", "page tab").click()
    xmleditor = win.find("XML editor")
    newtext = xmleditor.text.replace(">%s<" % vmname, ">%s<" % newname)
    xmleditor.set_text(newtext)
    win.find("Migrate", "push button").click()
    lib.utils.check(lambda: not win.showing, timeout=10)

    manager.find(newname, "table cell")

    # Do standard xmleditor tests
    win = _open_migrate(app, vmname)
    win.find("address-text").set_text("TESTSUITE-FAKE")
    finish = win.find("Migrate", "push button")
    lib.utils.test_xmleditor_interactions(app, win, finish)
    win.find("Cancel", "push button").click()
    lib.utils.check(lambda: not win.visible)



# Copyright 2017 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from ..xmlbuilder import XMLBuilder, XMLChildProperty, XMLProperty


XMLBuilder.register_namespace("libosinfo", "http://libosinfo.org/xmlns/libvirt/domain/1.0")


class _XMLNSLibosinfo(XMLBuilder):
    XML_NAME = "libosinfo:libosinfo"

    os_id = XMLProperty("./libosinfo:os/@id")


class DomainMetadata(XMLBuilder):
    """
    Class for generating <metadata> XML
    """

    XML_NAME = "metadata"

    libosinfo = XMLChildProperty(_XMLNSLibosinfo, is_single=True)



#
# Copyright 2010, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from ..xmlbuilder import XMLBuilder, XMLProperty


class DomainFeatures(XMLBuilder):
    """
    Class for generating <features> XML
    """

    XML_NAME = "features"
    _XML_PROP_ORDER = ["acpi", "apic", "pae", "gic_version"]

    acpi = XMLProperty("./acpi", is_bool=True)
    apic = XMLProperty("./apic", is_bool=True)
    pae = XMLProperty("./pae", is_bool=True)
    gic_version = XMLProperty("./gic/@version")

    hap = XMLProperty("./hap", is_bool=True)
    viridian = XMLProperty("./viridian", is_bool=True)
    privnet = XMLProperty("./privnet", is_bool=True)

    pmu = XMLProperty("./pmu/@state", is_onoff=True)
    eoi = XMLProperty("./apic/@eoi", is_onoff=True)

    hyperv_relaxed = XMLProperty("./hyperv/relaxed/@state", is_onoff=True)
    hyperv_vapic = XMLProperty("./hyperv/vapic/@state", is_onoff=True)
    hyperv_spinlocks = XMLProperty("./hyperv/spinlocks/@state", is_onoff=True)
    hyperv_spinlocks_retries = XMLProperty("./hyperv/spinlocks/@retries", is_int=True)
    hyperv_vpindex = XMLProperty("./hyperv/vpindex/@state", is_onoff=True)
    hyperv_runtime = XMLProperty("./hyperv/runtime/@state", is_onoff=True)
    hyperv_synic = XMLProperty("./hyperv/synic/@state", is_onoff=True)
    hyperv_stimer = XMLProperty("./hyperv/stimer/@state", is_onoff=True)
    hyperv_stimer_direct = XMLProperty("./hyperv/stimer/direct/@state", is_onoff=True)
    hyperv_reset = XMLProperty("./hyperv/reset/@state", is_onoff=True)
    hyperv_frequencies = XMLProperty("./hyperv/frequencies/@state", is_onoff=True)
    hyperv_reenlightenment = XMLProperty("./hyperv/reenlightenment/@state", is_onoff=True)
    hyperv_tlbflush = XMLProperty("./hyperv/tlbflush/@state", is_onoff=True)
    hyperv_tlbflush_direct = XMLProperty("./hyperv/tlbflush/direct/@state", is_onoff=True)
    hyperv_tlbflush_extended = XMLProperty("./hyperv/tlbflush/extended/@state", is_onoff=True)
    hyperv_ipi = XMLProperty("./hyperv/ipi/@state", is_onoff=True)
    hyperv_evmcs = XMLProperty("./hyperv/evmcs/@state", is_onoff=True)
    hyperv_avic = XMLProperty("./hyperv/avic/@state", is_onoff=True)
    hyperv_xmm_input = XMLProperty("./hyperv/xmm_input/@state", is_onoff=True)
    hyperv_emsr_bitmap = XMLProperty("./hyperv/emsr_bitmap/@state", is_onoff=True)

    vmport = XMLProperty("./vmport/@state", is_onoff=True)
    kvm_hidden = XMLProperty("./kvm/hidden/@state", is_onoff=True)
    kvm_hint_dedicated = XMLProperty("./kvm/hint-dedicated/@state", is_onoff=True)
    kvm_poll_control = XMLProperty("./kvm/poll-control/@state", is_onoff=True)
    kvm_pv_ipi = XMLProperty("./kvm/pv-ipi/@state", is_onoff=True)
    pvspinlock = XMLProperty("./pvspinlock/@state", is_onoff=True)

    smm = XMLProperty("./smm/@state", is_onoff=True)
    vmcoreinfo = XMLProperty("./vmcoreinfo/@state", is_onoff=True)
    ioapic_driver = XMLProperty("./ioapic/@driver")
    msrs_unknown = XMLProperty("./msrs/@unknown")

    ##################
    # Default config #
    ##################

    def _set_hyperv_defaults(self, guest):
        if not guest.hyperv_supported():
            return

        features = guest.lookup_domcaps().supported_hyperv_features()

        def _enable(name, requires=None, feature=None, value=True):
            feature = feature or name
            if feature not in features:
                return
            if getattr(self, f"hyperv_{name}") is not None:
                return
            if requires:
                for val in requires:
                    if getattr(self, f"hyperv_{val}") is not True:
                        return
            setattr(self, f"hyperv_{name}", value)

        _enable("relaxed")
        _enable("vapic")
        _enable("spinlocks")
        _enable("spinlocks_retries", feature="spinlocks", value=8191)
        _enable("vpindex")
        _enable("runtime")
        _enable("synic", requires=["vpindex"])

        # Both hyperv_stimer and hyperv_stimer requires hv-timer to be enabled
        # which libvirt hides under hypervclock timer.
        if guest.clock.has_hyperv_timer():
            _enable("stimer", requires=["vpindex", "synic"])
            _enable("stimer_direct", requires=["vpindex", "synic", "stimer"])

        _enable("frequencies")

        _enable("tlbflush", requires=["vpindex"])
        _enable("ipi", requires=["vpindex"])

        if guest.conn.caps.host.cpu.vendor == "Intel":
            _enable("evmcs", requires=["vapic"])

        if self.apic is True:
            _enable("avic")

    def set_defaults(self, guest):
        if guest.os.is_container():
            self.acpi = None
            self.apic = None
            self.pae = None
            if guest.is_full_os_container() and guest.type != "vz":
                self.privnet = True
            return

        if not guest.os.is_hvm():
            return

        capsinfo = guest.lookup_capsinfo()
        if self._prop_is_unset("acpi"):
            self.acpi = capsinfo.guest.supports_acpi()
        if self._prop_is_unset("apic"):
            self.apic = capsinfo.guest.supports_apic()
        if self._prop_is_unset("pae"):
            if guest.os.is_hvm() and guest.type == "xen" and guest.os.arch == "x86_64":
                self.pae = True
            else:
                self.pae = capsinfo.guest.supports_pae()

        self._set_hyperv_defaults(guest)



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import pytest

from . import lib

try:
    import guestfs  # pylint: disable=import-error

    ignore = guestfs
    HAS_LIBGUESTFS = True
except Exception:
    HAS_LIBGUESTFS = False


#########################################################
# UI tests for the libguestfs inspection infrastructure #
#########################################################


def testInspectionMock(app):
    if not HAS_LIBGUESTFS:
        pytest.skip("libguestfs python not installed")

    # Use the test suite inspection mocking to test parts
    # of the UI that interact with inspection data
    app.open(enable_libguestfs=True)
    manager = app.topwin

    details = app.manager_open_details("test-clone")
    details.find("OS information", "table cell").click()
    tab = details.find("os-tab")

    tab.find("Application", "toggle").click_expander()
    apps = tab.find("inspection-apps")
    apps.check_onscreen()
    apps.click_expander()

    nodestr1 = apps.fmt_nodes()
    assert "test_app1_summary" in nodestr1
    tab.find("Refresh", "push button").click()
    lib.utils.check(lambda: apps.fmt_nodes() != nodestr1)

    details.window_close()

    # Open a VM with no disks which will report an inspection error
    app.root.find_fuzzy("test\n", "table cell").doubleClick()
    details = app.find_window("test on")
    details.find("Details", "radio button").click()
    details.find("OS information", "table cell").click()
    tab = details.find("os-tab")
    tab.find_fuzzy("Fake test error no disks")

    # Closing and reopening a connection triggers some libguest
    # cache reading
    details.window_close()
    manager.click()
    c = manager.find("test testdriver.xml", "table cell")
    app.manager_conn_disconnect("test testdriver.xml")
    lib.utils.check(lambda: "Not Connected" in c.text)
    app.manager_conn_connect("test testdriver.xml")
    lib.utils.check(lambda: "Not Connected" not in c.text)



#
# Copyright 2010, 2012-2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from ..xmlbuilder import XMLBuilder, XMLProperty


class DomainSeclabel(XMLBuilder):
    """
    Class for generating <seclabel> XML
    """

    TYPE_DYNAMIC = "dynamic"
    TYPE_STATIC = "static"

    MODEL_TEST = "testSecurity"
    MODEL_SELINUX = "selinux"
    MODEL_DAC = "dac"
    MODEL_NONE = "none"

    XML_NAME = "seclabel"
    _XML_PROP_ORDER = ["type", "model", "relabel", "label"]

    def _guess_secmodel(self):
        caps_models = [x.model for x in self.conn.caps.host.secmodels]

        # We always want the testSecurity model when running tests
        if self.MODEL_TEST in caps_models:
            return self.MODEL_TEST
        if not self.label:
            return caps_models and caps_models[0] or None

        lab_len = None
        if self.label:
            lab_len = min(3, len(self.label.split(":")))

        if lab_len == 3:
            return self.MODEL_SELINUX
        elif lab_len == 2:
            return self.MODEL_DAC

    model = XMLProperty("./@model")
    type = XMLProperty("./@type")

    label = XMLProperty("./label")
    baselabel = XMLProperty("./baselabel")
    relabel = XMLProperty("./@relabel", is_yesno=True)

    ##################
    # Default config #
    ##################

    def set_defaults(self, _guest):
        if not self.type and not self.model:
            # Let libvirt fill it in
            return
        if self.type is None:
            self.type = self.TYPE_DYNAMIC
        if self.model is None:
            self.model = self._guess_secmodel()
